#include "fintrf.h"
C======================================================================
#if 0
C  This is mex file version of L1SamplerExample1.m that is usually faster than the matlab version
C  
C Written for:
C F.Lucka, Fast MCMC sampling for sparse Bayesian inference in high-dimensional 
C inverse problems using L1-type priors, submitted to Inverse Problems,
C 2012.
C
C	 Detailed comments on the code can be found in the corresponding .m file, this code is a direct
C reimplementation of the matlab code using mex.	
C
C I don't have much knowledge about Fortran and mex files, so the code is rather unelegant 
C 
C Written by: Felix Lucka, WWU Münster, Germany
C Email: felix.lucka@wwu.de
C Created: May 2012
C
#endif
C     
C======================================================================
C     Gateway routine
      subroutine mexFunction(nlhs, plhs, nrhs, prhs)

C     Declarations
      implicit none

C     mexFunction arguments:
      mwPointer plhs(*), prhs(*)
      integer nlhs, nrhs

C     Function declarations:
      mwPointer mxGetPr
      mwPointer mxCreateDoubleMatrix
      integer mxIsNumeric
      integer*4, external :: mexPrintf, mexEvalString
      real*8 L1SamplerFunc, DoubleExpSampler
      mwSize mxGetM, mxGetN

C     Pointers to input/output mxArrays:
      mwPointer Psi_ptr, m_bar_ptr, c_vec_ptr, K0_ptr, K_ptr
      mwPointer thinning_ptr, N_O_ptr, rn_scan_fl_ptr, chain_fun_ptr
      mwPointer zhi_chain_ptr, zhi_chain_ind_ptr

C     Array information:
      mwSize mrows, ncols

C     (Almost) everything else
		real*8, allocatable, dimension(:,:) :: Psi, Psi_tr,zhi_chain2
		real*8, allocatable, dimension(:,:) :: Psi_t_Psi
		real*8, allocatable, dimension(:) :: m_bar, c_vec, zhi, psi_norm
		real*8, allocatable, dimension(:) :: project_v, zhi_chain1
		real*8, allocatable, dimension(:) :: zhi_chain_ind
		real*8, allocatable, dimension(:) :: all_sqrt_a, Psi_i_m_bar
		real*8, allocatable, dimension(:) :: unit_rn
		integer*4, allocatable, dimension(:) :: rn_comp_ind
      real*8  :: dummy_real, sqrt_a, b, c, x_curr, y, a1,a2,a3
      integer*8 :: K0, K, K_tot, thinning, chain_index, i81
      integer*8 :: j81, next_output_step, output_inc
      integer*4 :: n, N_O, m_chain, i41, k_print, i_comp
      logical  :: rn_scan_fl, fun_fl, OOR_fl, n_max_fl
      character(len=80) :: line

C		Parameter
C		the number of unknowns up to which Phi is build and used (cf. eq 19)
		integer*4, parameter :: n_max_Phi_impl = 20000
      
C-----------------------------------------------------------------------
C     Check for proper number of arguments. 
      if(nrhs .ne. 9) then
         call mexErrMsgIdAndTxt ('MATLAB:L1SCsampler:nInput',
     +                           'Nine inputs required.')
      elseif(nlhs .ne. 2) then
         call mexErrMsgIdAndTxt ('MATLAB:L1SCsampler:nOutput',
     +                           'Two output arguments required.')
      endif


C     Create Fortran arrays for the input arguments.
C		Psi
      Psi_ptr = mxGetPr(prhs(1))
		mrows = mxGetM(prhs(1))
      ncols = mxGetN(prhs(1))
      allocate(Psi(mrows,ncols))
      call mxCopyPtrToReal8(Psi_ptr,Psi,mrows*ncols)
		n = size(Psi,2)
		
C		m_bar      
      m_bar_ptr = mxGetPr(prhs(2))
		mrows = mxGetM(prhs(2))
      allocate(m_bar(mrows))
      call mxCopyPtrToReal8(m_bar_ptr,m_bar,mrows)

C		c_vec      
      c_vec_ptr = mxGetPr(prhs(3))
		mrows = mxGetM(prhs(3))
      allocate(c_vec(mrows))
      call mxCopyPtrToReal8(c_vec_ptr,c_vec,mrows)

C		K0, K, thinning, N_O
      K0_ptr = mxGetPr(prhs(4))
		call mxCopyPtrToReal8(K0_ptr,dummy_real,1)
		K0 = int(dummy_real,8)
      K_ptr = mxGetPr(prhs(5))
      	call mxCopyPtrToReal8(K_ptr,dummy_real,1)
      	K = int(dummy_real,8)
      thinning_ptr = mxGetPr(prhs(6))
     	call mxCopyPtrToReal8(thinning_ptr,dummy_real,1)
      	thinning = int(dummy_real,8)
      N_O_ptr = mxGetPr(prhs(7))
      call mxCopyPtrToReal8(N_O_ptr,dummy_real,1)
      N_O = int(dummy_real,4)

C		rn_scan_fl     
      rn_scan_fl_ptr = mxGetPr(prhs(8))
     	call mxCopyPtrToReal8(rn_scan_fl_ptr,dummy_real,1)
		rn_scan_fl = dummy_real > 0
		if(rn_scan_fl) then
			allocate(rn_comp_ind(n))
			allocate(unit_rn(n))
		end if

C		if chain_fun is a scalar, store every state, if it is a vector of size n,
C 	project every state onto it
      chain_fun_ptr = mxGetPr(prhs(9))
      	mrows = mxGetM(prhs(9))
      ncols = mxGetN(prhs(9))
      if(mrows == ncols .and. mrows == 1) then
      		fun_fl = .false.
      else if (mrows == n .and. ncols == 1) then
      		fun_fl  = .true.
      		allocate(project_v(n))
			call mxCopyPtrToReal8(chain_fun_ptr,project_v,n)
		else
         call mexErrMsgIdAndTxt ('MATLAB:L1SamplerExample1Mex:nOutput',
     +					     'Invalid input for chain_fun_fl.')
      end if


C		 Use ordered overrelaxation ?
		OOR_fl = N_O > 1
		if(OOR_fl) then
    		if(mod(N_O,2) /= 1) then
         call mexErrMsgIdAndTxt ('MATLAB:L1SamplerExample1Mex:nOutput',
     +					     'N_O has to be odd!')
    		end if
		end if


            
   		allocate(zhi(n))
		zhi = 0.0

C allocate chains
		K_tot = K+K0
		a1 = real(K_tot)
		a2 = real(K0)
		a3 = real(thinning)
		m_chain = int(floor(a1/a3)-floor(a2/a3),4)
		if(fun_fl) then
			allocate(zhi_chain1(m_chain))
			zhi_chain1 = 0.0
			plhs(1) = mxCreateDoubleMatrix(1,m_chain,0)
			zhi_chain_ptr = mxGetPr(plhs(1))
		else
			allocate(zhi_chain2(n,m_chain))
			zhi_chain2 = 0.0
			plhs(1) = mxCreateDoubleMatrix(n,m_chain,0)
			zhi_chain_ptr = mxGetPr(plhs(1))
		end if
		
		allocate(zhi_chain_ind(m_chain))
		zhi_chain_ind  = 0.0
		chain_index = 1
		plhs(2) = mxCreateDoubleMatrix(1,m_chain,0.0)
		zhi_chain_ind_ptr = mxGetPr(plhs(2))


C Initialize some stuff for the sampling (cf. Section 2.3.3.)
		allocate(psi_norm(n),all_sqrt_a(n),Psi_i_m_bar(n))
		do i41=1,n
    		psi_norm(i41) = dot_product(Psi(:,i41),Psi(:,i41))
    		Psi_i_m_bar(i41) = dot_product(Psi(:,i41), m_bar)
		end do
		all_sqrt_a = sqrt(psi_norm)
		

C the number of unknowns up to which Phi is build and used (cf. eq 19)
		
		n_max_fl = n < n_max_Phi_impl
		if(n_max_fl) then
			allocate(Psi_t_Psi(n,n))
    		Psi_t_Psi =  matmul(transpose(Psi), Psi )
    		deallocate(Psi)
    	else
    		allocate(Psi_tr(ncols,mrows))
    		Psi_tr = transpose(Psi)
		end if

C display some output
		write(line,*) "Compute CM estimate via L1 single component"
		k_print = mexPrintf(line)
		write(line,*) "Gibbs sampler using "
		k_print = mexPrintf(line)
		if(rn_scan_fl) then
			write(line,*) "random scanning"
			k_print = mexPrintf(line)
		else
			write(line,*) "systematic scanning"
			k_print = mexPrintf(line)
		end if
		if(OOR_fl) then
			write(line,*) "and ordered overrelaxation with N_O = ",N_O
			k_print = mexPrintf(line)
		end if
		write(line,*) "\b."
		k_print = mexPrintf(line//achar(10))

		write(line,*) "Full status bar for comparison:"
		k_print = mexPrintf(line//achar(10))
		a1 = real(K_tot)
		a2 = a1/min(100.0,a1)
		next_output_step = int(floor(a2),8)
		output_inc = next_output_step
		do j81 = 1,K_tot
			if(j81 == next_output_step) then
				write(line,*) "\b."
				k_print = mexPrintf(line)
				next_output_step = next_output_step + output_inc
			end if
		end do
		next_output_step = output_inc
		write(line,*) "\nCurrent status bar:"
		k_print = mexPrintf(line//achar(10))
		k_print = mexEvalString("drawnow;")

		do j81 = 1,K_tot
			if(rn_scan_fl) then
				call random_number(unit_rn)	
         		rn_comp_ind = ceiling(real(n) * unit_rn,4)
    		end if

    		do i41=1,n
    			if(rn_scan_fl) then
      		      i_comp = rn_comp_ind(i41)
      		  else
      		      i_comp = i41
      		  end if

      		  if(psi_norm(i_comp) > 0.0) then
C				if a likelyhood term is present, sample from the 1 dim
C				densiy from exp(-a*x^2+b*x+c*abs(x)) via inverse cumulative
C           distibution method      		  
                sqrt_a = all_sqrt_a(i_comp)
                if(n_max_fl) then
                	b = Psi_i_m_bar(i_comp)
                	b = b - dot_product(zhi,Psi_t_Psi(:,i_comp))
                	b = b + psi_norm(i_comp)*zhi(i_comp)
                	b = 2.0*b
                else
                	b = Psi_i_m_bar(i_comp)
                	b = b - dot_product(matmul(Psi_tr,Psi(:,i_comp)),zhi) 
                	b = b + psi_norm(i_comp)*zhi(i_comp)
                	b = 2.0*b
                end if
                c = c_vec(i_comp)  
                x_curr = zhi(i_comp)
					 y = L1SamplerFunc (sqrt_a,b,c,N_O,x_curr)
      		  else
C     	  	if no likelyhood term is present, just draw from the two
C         	sided exponential distribution by inverse cumulative
C         	distribution method.
					y = DoubleExpSampler(c_vec(i_comp),N_O,zhi(i_comp))
      		  end if
				zhi(i_comp) = y
    		end do

C 		Store current state
    		if(j81 > K0 .and. mod(j81,thinning) == 0) then
         		if(fun_fl) then
               zhi_chain1(chain_index) = dot_product(project_v,zhi)
            else
               zhi_chain2(:,chain_index) = zhi
            end if
        		zhi_chain_ind(chain_index) = real(j81,8)
        		chain_index = chain_index + 1
    		end if
    
			if(j81 == next_output_step) then
				write(line,*) "\b."
				k_print = mexPrintf(line)
				k_print = mexEvalString("drawnow;")
				next_output_step = next_output_step + output_inc
			end if
		end do
		write(line,*) "\n"
		k_print = mexPrintf(line)
		
C		deallocate stuff
		if(n_max_fl) then
			deallocate(Psi_t_Psi)
		else
			deallocate(Psi,Psi_tr)
		end if
		deallocate (m_bar,c_vec)
		if(fun_fl) then
	 		deallocate(project_v)
		end if
		deallocate(zhi,all_sqrt_a,psi_norm,Psi_i_m_bar)
		if(rn_scan_fl) then
			deallocate(rn_comp_ind,unit_rn)
		end if

C     Load the data into chain_ptr, which is the output to MATLAB.
    	i81 = size(zhi_chain_ind,1) 
		call mxCopyReal8ToPtr(zhi_chain_ind,zhi_chain_ind_ptr,i81)     
		deallocate (zhi_chain_ind)
		
		if(fun_fl) then
			i81 = size(zhi_chain1,1)
			call mxCopyReal8ToPtr(zhi_chain1,zhi_chain_ptr,i81)     
			deallocate(zhi_chain1)
		else
			i81 = size(zhi_chain2,1) * size(zhi_chain2,2)
			call mxCopyReal8ToPtr(zhi_chain2,zhi_chain_ptr,i81)     
			deallocate(zhi_chain2)
		end if

      return
      end


