#include "fintrf.h"
C======================================================================
#if 0
C  This is mex file version of L1SamplerExample2.m that is usually faster than the matlab version
C  
C Written for:
C F.Lucka, Fast MCMC sampling for sparse Bayesian inference in high-dimensional 
C inverse problems using L1-type priors, submitted to Inverse Problems,
C 2012.
C
C	 Detailed comments on the code can be found in the corresponding .m file, this code is a direct
C reimplementation of the matlab code using mex.	
C
C I don't have much knowledge about Fortran and mex files, so the code is rather unelegant 
C 
C Written by: Felix Lucka, WWU Münster, Germany
C Email: felix.lucka@wwu.de
C Created: May 2012
C
C   
#endif
C     
C======================================================================
C     Gateway routine
      subroutine mexFunction(nlhs, plhs, nrhs, prhs)

C     Declarations
      implicit none

C     mexFunction arguments:
      mwPointer plhs(*), prhs(*), plhs_c(1), prhs_c(3)
      integer nlhs, nrhs

C     Function declarations:
      mwPointer mxGetPr
      mwPointer mxCreateDoubleMatrix, mxCreateNumericArray
      integer mxIsNumeric
      integer*4, external :: mexPrintf, mexEvalString
      integer*4, external ::  mxClassIDFromClassName
      integer*4, external :: mexCallMATLAB, mxCreateString
C      real*8 :: L1SamplerFunc, DoubleExpSampler
      real*8 :: pad_array
      mwSize mxGetM, mxGetN

C		Interfaces
		Interface
			function crop_mat_ind(A)
			real*8 :: A(:,:)
			integer*4 :: crop_mat_ind(4)
			end function crop_mat_ind

			function pad_array(A,padsize)
			real*8 :: A(:,:)
			real*8, allocatable, dimension(:,:) :: pad_array
			integer*4 :: padsize(2)
			end function pad_array

			function L1SamplerFunc ( sqrt_a,b,c,N_OOR,x_curr)
			real*8 sqrt_a, b, c, x_curr, L1SamplerFunc 
			integer*4 N_OOR
			end function L1SamplerFunc

			function DoubleExpSampler (c,N_O,x_curr)
			integer*4 N_O
			real*8 :: c, x_curr, DoubleExpSampler
			end function DoubleExpSampler 
		end Interface


C     Pointers to input/output mxArrays:
      mwPointer Psi_psf_ptr, PtP_psf_ptr, m_bar_ptr, c_im_ptr
      mwPointer K0_ptr, K_ptr, thinning_ptr, N_O_ptr,out_fl_ptr
      mwPointer rn_scan_fl_ptr, zhi_chain_ptr, zhi_chain_ind_ptr

C     Array information:
      mwSize mrows, ncols

C     (Almost) everything else
		real*8, allocatable, dimension(:,:) :: Psi_psf, Psi_psf_c,m_bar
		real*8, allocatable, dimension(:,:) :: PtP_psf, PtP_psf_c
		real*8, allocatable, dimension(:,:) :: c_im, zhi, psi_norm
		real*8, allocatable, dimension(:,:,:) ::  zhi_chain
		real*8, allocatable, dimension(:) :: zhi_chain_ind
		real*8, allocatable, dimension(:,:) :: all_sqrt_a, Psi_i_m_bar
		real*8, allocatable, dimension(:,:) :: aux_arr,template
		 real*8, allocatable, dimension(:,:) :: m_bar_pad, zhi_pad
		real*8, allocatable, dimension(:) :: unit_rn
		integer*4, allocatable, dimension(:) :: rn_comp_ind
		integer*4, allocatable, dimension(:) ::  xi_P,yi_P
		integer*4, allocatable, dimension(:) ::  xi_PtP,yi_PtP
		integer*4, allocatable, dimension(:,:) :: si
      real*8  :: dummy_real, sqrt_a, b, c, x_curr, y, a1,a2,a3
      integer*8 :: K0, K, K_tot, thinning, chain_index, i81
      integer*8 :: j81, next_output_step, output_inc
      integer*4 :: m,n, dof, N_O, m_chain, i4, j4,k_mat, i_comp
      integer*4 :: m_Psi_c,n_Psi_c,m_PtP_c,n_PtP_c
      integer*4 :: classid, dims(3),cut_ind(4),padsize(2),ics,jcs
      integer*4 :: x_os_PtP,y_os_PtP
      logical  :: rn_scan_fl, fun_fl, OOR_fl, n_max_fl, not_ready_fl
      logical  :: out_fl
      character(len=80) :: line

      
C-----------------------------------------------------------------------
C     Check for proper number of arguments. 
      if(nrhs .ne. 10) then
         call mexErrMsgIdAndTxt ('MATLAB:L1SCsampler:nInput',
     +                           'ten inputs required.')
      elseif(nlhs .ne. 2) then
         call mexErrMsgIdAndTxt ('MATLAB:L1SCsampler:nOutput',
     +                           'Two output arguments required.')
      endif


C     Create Fortran arrays for the input arguments.
C		Psi_psf
      Psi_psf_ptr = mxGetPr(prhs(1))
		mrows = mxGetM(prhs(1))
      ncols = mxGetN(prhs(1))
      allocate(Psi_psf(mrows,ncols))
      call mxCopyPtrToReal8(Psi_psf_ptr,Psi_psf,mrows*ncols)
		m = size(Psi_psf,1)
		n = size(Psi_psf,2)
		dof = m*n
     
C		Psi_tr_Psi_psf 
      PtP_psf_ptr = mxGetPr(prhs(2))
		mrows = mxGetM(prhs(2))
      ncols = mxGetN(prhs(2))
      allocate(PtP_psf(mrows,ncols))
      call mxCopyPtrToReal8(PtP_psf_ptr,PtP_psf,mrows*ncols)
      
C 	make it symmetric again
		PtP_psf = (PtP_psf + transpose(PtP_psf) )/2.0
C 	Cut off too small values
		a1 = epsilon(dummy_real)
	 	a2 =  a1 * maxval(abs(Psi_psf))
	 	where(abs(Psi_psf) < a2) Psi_psf = 0.0
	 	where(abs(PtP_psf) < a2) PtP_psf = 0.0
	 	
C		m_bar      
      m_bar_ptr = mxGetPr(prhs(3))
      mrows = mxGetM(prhs(3))
      ncols = mxGetN(prhs(3))
      allocate(m_bar(mrows,ncols))
      call mxCopyPtrToReal8(m_bar_ptr,m_bar,mrows*ncols)
		
C		c_im      
      c_im_ptr = mxGetPr(prhs(4))
		mrows = mxGetM(prhs(4))
		ncols = mxGetN(prhs(4))
      allocate(c_im(mrows,ncols))
      call mxCopyPtrToReal8(c_im_ptr,c_im,mrows*ncols)

C		K0, K, thinning, N_O
      K0_ptr = mxGetPr(prhs(5))
		call mxCopyPtrToReal8(K0_ptr,dummy_real,1)
		K0 = int(dummy_real,8)
      K_ptr = mxGetPr(prhs(6))
      	call mxCopyPtrToReal8(K_ptr,dummy_real,1)
      	K = int(dummy_real,8)
      thinning_ptr = mxGetPr(prhs(7))
     	call mxCopyPtrToReal8(thinning_ptr,dummy_real,1)
      	thinning = int(dummy_real,8)
      N_O_ptr = mxGetPr(prhs(8))
      call mxCopyPtrToReal8(N_O_ptr,dummy_real,1)
      N_O = int(dummy_real,4)

C		rn_scan_fl     
      rn_scan_fl_ptr = mxGetPr(prhs(9))
     	call mxCopyPtrToReal8(rn_scan_fl_ptr,dummy_real,1)
		rn_scan_fl = dummy_real > 0
		if(rn_scan_fl) then
			allocate(rn_comp_ind(dof),unit_rn(dof))
		end if
C		output flag     
      out_fl_ptr = mxGetPr(prhs(10))
     	call mxCopyPtrToReal8(out_fl_ptr,dummy_real,1)
		out_fl = dummy_real > 0


C		 Use ordered overrelaxation ?
		OOR_fl = N_O > 1
		if(OOR_fl) then
    		if(mod(N_O,2) /= 1) then
         call mexErrMsgIdAndTxt ('MATLAB:L1SamplerExample1Mex:nOutput',
     +					     'N_O has to be odd!')
    		end if
		end if
            
   		allocate(zhi(m,n))
		zhi = 0.0
     
C allocate chains
		K_tot = K+K0
		a1 = real(K_tot)
		a2 = real(K0)
		a3 = real(thinning)
		m_chain = int(floor(a1/a3)-floor(a2/a3),4)

		allocate(zhi_chain(m,n,m_chain))
		zhi_chain = 0.0
		dims = (/ m,n,m_chain /)  
		classid = mxClassIDFromClassName("double")
		plhs(1) = mxCreateNumericArray(3, dims, classid,0)
		zhi_chain_ptr = mxGetPr(plhs(1))
		
		allocate(zhi_chain_ind(m_chain))
		zhi_chain_ind  = 0.0
		chain_index = 1
		plhs(2) = mxCreateDoubleMatrix(1,m_chain,0.0)
		zhi_chain_ind_ptr = mxGetPr(plhs(2))


C Initialize some stuff for the sampling (cf. Section 2.3.3.)
		allocate(psi_norm(m,n),all_sqrt_a(m,n),Psi_i_m_bar(m,n))
		psi_norm = 0.0
		Psi_i_m_bar = 0.0

		cut_ind = crop_mat_ind(Psi_psf)
		m_Psi_c = cut_ind(2)-cut_ind(1)+1
		n_Psi_c = cut_ind(4)-cut_ind(3)+1
		allocate(Psi_psf_c(m_Psi_c,n_Psi_c))
		Psi_psf_c = Psi_psf(cut_ind(1):cut_ind(2),cut_ind(3):cut_ind(4))
		allocate(xi_P (m_Psi_c),yi_P(n_Psi_c))
		xi_P = (/(i4,i4=0,m_Psi_c-1)/) 
		yi_P = (/(i4,i4=0,n_Psi_c-1)/) 
		padsize(1) = floor(m_Psi_c/2.0)
		padsize(2) = floor(n_Psi_c/2.0)		
		allocate(template(m+2*padsize(1),n+2*padsize(2)),aux_arr(m,n))
		allocate(m_bar_pad(m+2*padsize(1),n+2*padsize(2)))
		aux_arr = 1.0

		template = pad_array(aux_arr,padsize)
		m_bar_pad = pad_array(m_bar,padsize)
		
		do i4=1,m
			do j4=1,n
	psi_norm(i4,j4)=sum((template(i4+xi_P,j4+yi_P)*Psi_psf_c)**2)
	Psi_i_m_bar(i4,j4)=sum((m_bar_pad(i4+xi_P,j4+yi_P)*Psi_psf_c))
			end do
		end do

		all_sqrt_a = sqrt(psi_norm)
		deallocate( template,m_bar_pad, aux_arr)

		cut_ind = crop_mat_ind(PtP_psf)
		m_PtP_c = cut_ind(2)-cut_ind(1)+1
		n_PtP_c = cut_ind(4)-cut_ind(3)+1
		allocate(PtP_psf_c(m_PtP_c,n_PtP_c))
		PtP_psf_c = PtP_psf(cut_ind(1):cut_ind(2),cut_ind(3):cut_ind(4))
		allocate(xi_PtP (m_PtP_c),yi_PtP(n_PtP_c))
		xi_PtP = (/(i4,i4=0,m_PtP_c-1)/) 
		yi_PtP = (/(i4,i4=0,n_PtP_c-1)/) 
		x_os_PtP =  floor(m_PtP_c/2.0)
		y_os_PtP =  floor(n_PtP_c/2.0)
		
		padsize(1) = floor(m_PtP_c/2.0)
		padsize(2) = floor(n_PtP_c/2.0)
		allocate(zhi_pad(m+2*padsize(1),n+2*padsize(2)))
		zhi_pad = 0.0

		allocate(si(dof,2))
		i81 = 1
		do j4=1,n
			do i4=1,m
				si(i81,1) = i4
				si(i81,2) = j4
				i81 = i81 + 1
			end do
		end do

		
C display some output
        if(out_fl) then
		write(line,*) "Compute CM estimate via L1 single component"
		k_mat = mexPrintf(line)
		write(line,*) "Gibbs sampler using "
		k_mat = mexPrintf(line)
		if(rn_scan_fl) then
			write(line,*) "random scanning"
			k_mat = mexPrintf(line)
		else
			write(line,*) "systematic scanning"
			k_mat = mexPrintf(line)
		end if
		if(OOR_fl) then
			write(line,*) "and ordered overrelaxation with N_O = ",N_O
			k_mat = mexPrintf(line)
		end if
		write(line,*) "\b."
		k_mat = mexPrintf(line//achar(10))

		write(line,*) "Full status bar for comparison:"
		k_mat = mexPrintf(line//achar(10))
		a1 = real(K_tot)
		a2 = a1/min(100.0,a1)
		next_output_step = int(floor(a2),8)
		output_inc = next_output_step
		do j81 = 1,K_tot
			if(j81 == next_output_step) then
				write(line,*) "\b."
				k_mat = mexPrintf(line)
				next_output_step = next_output_step + output_inc
			end if
		end do
		next_output_step = output_inc
		write(line,*) "\nCurrent status bar:"
		k_mat = mexPrintf(line//achar(10))
		k_mat = mexEvalString("drawnow")
        end if

     
		do j81 = 1,K_tot
			if(rn_scan_fl) then
				call random_number(unit_rn)	
         		rn_comp_ind = ceiling(real(dof) * unit_rn,4)
    		end if

    		do i4=1,dof
    			if(rn_scan_fl) then
      		      i_comp = rn_comp_ind(i4)
      		  else
      		      i_comp = i4
      		  end if

            ics = si(i_comp,1)
            jcs = si(i_comp,2)

      		  if(psi_norm(ics,jcs) > 0.0) then
C				if a likelyhood term is present, sample from the 1 dim
C				densiy from exp(-a*x^2+b*x+c*abs(x)) via inverse cumulative
C           distibution method      		  
               sqrt_a = all_sqrt_a(ics,jcs)
					
	            b = sum(zhi_pad(ics+xi_PtP,jcs+yi_PtP) * PtP_psf_c)
           	   b = Psi_i_m_bar(ics,jcs)-b+psi_norm(ics,jcs)*zhi(ics,jcs)
               b = 2*b
               c = c_im(ics,jcs)  
               x_curr = zhi(ics,jcs)
 					y = L1SamplerFunc (sqrt_a,b,c,N_O,x_curr)
      		  else
C     	  	if no likelyhood term is present, just draw from the two
C         	sided exponential distribution by inverse cumulative
C         	distribution method.
				y = DoubleExpSampler(c_im(ics,jcs),N_O,zhi(ics,jcs))
      		  end if
					zhi(ics,jcs) = y
	       		zhi_pad(ics+x_os_PtP,jcs+y_os_PtP) = y
    		end do

C 		Store current state
    		if(j81 > K0 .and. mod(j81,thinning) == 0) then
             zhi_chain(:,:,chain_index) = zhi
        		zhi_chain_ind(chain_index) = real(j81,8)
        		chain_index = chain_index + 1
    		end if
        
			if(out_fl .and. j81 == next_output_step) then
				write(line,*) "\b."
				k_mat = mexPrintf(line)
				k_mat = mexEvalString("drawnow")
				next_output_step = next_output_step + output_inc
			end if
		end do
        if(out_fl) then
		write(line,*) "\n"
		k_mat = mexPrintf(line)
        end if

C		deallocate stuff

		deallocate(Psi_psf,Psi_psf_c,PtP_psf,PtP_psf_c)
		deallocate (m_bar,c_im)
		deallocate(zhi,zhi_pad,all_sqrt_a,psi_norm,Psi_i_m_bar)
		deallocate(xi_P ,yi_P,xi_PtP ,yi_PtP,si)
		if(rn_scan_fl) then
			deallocate(rn_comp_ind,unit_rn)
		end if
		
C     Load the data into chain_ptr, which is the output to MATLAB.
    	i81 = size(zhi_chain_ind,1) 
		call mxCopyReal8ToPtr(zhi_chain_ind,zhi_chain_ind_ptr,i81)     
		deallocate (zhi_chain_ind)
		

		i81 = size(zhi_chain,3)
		call mxCopyReal8ToPtr(zhi_chain,zhi_chain_ptr,m*n*i81)     
		deallocate(zhi_chain)


      return
      end


