!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This is a collection of Fortran routines used by the mex files that I have written to speed up the 
! illustration of the results of: 
!
! Written for:
! F.Lucka, Fast MCMC sampling for sparse Bayesian inference in high-dimensional 
! inverse problems using L1-type priors, submitted to Inverse Problems,
! 2012.
!
! I don't have much knowledge about Fortran, so the code is rather unelegant 
! 
! Written by: Felix Lucka, WWU Münster, Germany
! Email: felix.lucka@wwu.de
! Created: May 2012
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

function L1SamplerFunc ( sqrt_a,b,c,N_OOR,x_curr)

implicit none

integer, parameter :: dp = kind(0.0d0)
! input/output
real(kind=dp) sqrt_a, b, c, x_curr, L1SamplerFunc 
! functions
real(kind=dp) myerfcx, myerfcinv, erfcinvexp_app, oor_unit_dist
! inner variables 1
real(kind=dp) alpha_plus, alpha_minus, gamma_
! inner variables 2
real(kind=dp) log_curr_prob, erfcx_curr, curr_prob, drawn_prob, arg_erfcx
! inner variables 3
real(kind=dp) u, z, y
! inner variables 4
real(kind=dp), parameter :: log_2 = 0.693147180559945, log_pi = 1.144729885849400
integer*4 N_OOR
logical OOR_fl


OOR_fl = (N_OOR > 1)

! determine alpha_+ and alpha_-
alpha_plus = (b+c)/(2.0*sqrt_a)
alpha_minus = (-b+c)/(2.0*sqrt_a)

! dependent on the case, use different fomula, since c > 0, it
! is not possible that both alpha_+ and alpha_- are negative

!!! the easy case, use fomua 1 !!!
if(alpha_plus > 0 .and. alpha_minus > 0) then
	gamma_ = myerfcx(alpha_plus) + myerfcx(alpha_minus)

	! Draw a new random number in [0,1]
	if(OOR_fl) then !Use oriented over relaxation
		! Compute current probability
		if(x_curr < 0) then
			erfcx_curr = myerfcx((- sqrt_a * x_curr + alpha_plus))
			log_curr_prob = -sqrt_a * x_curr * (sqrt_a * x_curr - 2.0 * alpha_plus) &
				+ log(erfcx_curr) - log(gamma_)
			curr_prob = exp(log_curr_prob)
		else
			erfcx_curr = myerfcx((sqrt_a * x_curr + alpha_minus))
			log_curr_prob = -sqrt_a * x_curr * (sqrt_a * x_curr + 2.0 * alpha_minus) &
				+ log(erfcx_curr) - log(gamma_)
			curr_prob = 1.0 - exp(log_curr_prob)
		end if
		drawn_prob = oor_unit_dist(curr_prob,N_OOR)
	else ! normal Gibbs sampler
		call random_number(drawn_prob)
	end if

	!! Part 1, check left side of distribution
	u = log(drawn_prob) - alpha_plus**2 + log(gamma_)
	if(u < -680.0 .and. abs(log(alpha_plus)) > 3.0) then
		z = erfcinvexp_app(u)
	else if(abs(log(alpha_plus)) > 3.0) then
		z = myerfcinv(exp(u))
	else
		z = myerfcinv(drawn_prob * exp(-alpha_plus**2) * gamma_)
	end if
	y = -(z - alpha_plus)/sqrt_a
	
	!! Part 2, check right side of distribution if y > 0
	if(y > 0.0) then
		u = log(1.0-drawn_prob) - alpha_minus**2 + log(gamma_)
		if(u < -680.0 .and. abs(log(alpha_minus)) > 3.0) then
			z = erfcinvexp_app(u)
		else if(abs(log(alpha_minus)) > 3.0) then
			z = myerfcinv(exp(u))
		else
			z = myerfcinv((1.0-drawn_prob) * exp(-alpha_minus**2) * gamma_)
		end if
		y =  (z - alpha_minus)/sqrt_a
	end if

!!! use fomula 2 !!!
else if(alpha_plus < 0) then
	gamma_ = myerfcx(-alpha_plus) - myerfcx(alpha_minus)

	! Draw a new random number in [0,1]
	if(OOR_fl) then ! Use oriented over relaxation
		! Compute current probability
		if(x_curr < 0) then
			arg_erfcx = - sqrt_a * x_curr + alpha_plus
			if(arg_erfcx > 0) then
				erfcx_curr = myerfcx(arg_erfcx)
				log_curr_prob = - arg_erfcx**2 + log(erfcx_curr) - log_2 &
					- log(1.0 - sign(1.0_8,gamma_) * exp( -alpha_plus**2 + log(abs(gamma_)) - log_2))
			else
				erfcx_curr = myerfcx(-arg_erfcx)
				log_curr_prob =  log(1.0 - exp(log(erfcx_curr) - log_2 - arg_erfcx**2)) &
					- log(1.0 - sign(1.0_8,gamma_) * exp( -alpha_plus**2 + log(abs(gamma_)) - log_2))
			end if
			curr_prob = exp(log_curr_prob)
		else
			arg_erfcx = sqrt_a * x_curr + alpha_minus
			if(arg_erfcx > 0) then
				erfcx_curr = myerfcx(arg_erfcx)
				log_curr_prob = - arg_erfcx**2 + log(erfcx_curr) - log_2 - (b*c)/sqrt_a**2 &
					- log(1.0 - sign(1.0_8,gamma_) * exp(-alpha_minus**2 + log(abs(gamma_)) - log_2 - (b*c)/sqrt_a**2))
			else
				erfcx_curr = myerfcx(-arg_erfcx)
				log_curr_prob = log(1.0 - exp(log(erfcx_curr) &
					- log_2 - arg_erfcx**2)) - (b*c)/sqrt_a**2 &
					- log(1.0 - sign(1.0_8,gamma_) * exp(-alpha_minus**2 + log(abs(gamma_)) - log_2 - (b*c)/sqrt_a**2))
			end if
			curr_prob = 1.0 - exp(log_curr_prob)
		end if
		drawn_prob = oor_unit_dist(curr_prob,N_OOR)
	else ! normal Gibbs sampler
		call random_number(drawn_prob)
	end if

	!! Part 1, check left side of distribution
	u = log(drawn_prob) + log_2 + &
			log(1.0 - sign(1.0_8,gamma_) * exp( -alpha_plus**2 + log(abs(gamma_)) - log_2))
	if(u < -680.0 .and. abs(log(-alpha_plus)) > 3.0) then
		z = erfcinvexp_app(u)
	else if(abs(log(-alpha_plus)) > 3.0) then
		z = myerfcinv(exp(u))
	else
		z = myerfcinv(2.0*drawn_prob - drawn_prob * exp(-alpha_plus**2) * gamma_)
	end if
	y = -(z - alpha_plus)/sqrt_a
	
	!! Part 2, check right side of distribution if y > 0
	if(y > 0.0) then
		u = log(1.0-drawn_prob) + log_2 + (b*c)/sqrt_a**2 + &
				log(1.0 - sign(1.0_8,gamma_)  * exp(-alpha_minus**2 + log(abs(gamma_)) - log_2 - (b*c)/sqrt_a**2))

		if(u < -680.0 .and. max( abs(log(-alpha_plus)) , abs(log(alpha_minus)) ) > 3.0) then
			z = erfcinvexp_app(u)
		else if(max( abs(log(-alpha_plus)) , abs(log(alpha_minus)) ) > 3.0) then
			z = myerfcinv(exp(u))
		else
			z = myerfcinv((1.0-drawn_prob) * (2.0*exp((b*c)/sqrt_a**2)  - exp(-alpha_minus**2) * gamma_))
		end if
		y =  (z - alpha_minus)/sqrt_a
	end if

!!! now alpha_minus is < 0, use fomula 3 !!!
else
	gamma_ = myerfcx(alpha_plus) - myerfcx(-alpha_minus)
	
	! Draw a new random number in [0,1]
	if(OOR_fl) then ! Use oriented over relaxation
		! Compute current probability
		if(x_curr < 0.0) then
			arg_erfcx = - sqrt_a * x_curr + alpha_plus
			erfcx_curr = myerfcx(arg_erfcx)   
			log_curr_prob = - (-sqrt_a * x_curr + alpha_plus)**2 + &
											log(erfcx_curr) - log_2 + (b*c)/sqrt_a**2 - &
											log(1.0 + sign(1.0_8,gamma_) * exp(-alpha_plus**2 + log(abs(gamma_)) - log_2 + (b*c)/sqrt_a**2 ))
			curr_prob = exp(log_curr_prob)
		else
			arg_erfcx = sqrt_a * x_curr + alpha_minus
			if(arg_erfcx > 0) then
				erfcx_curr = myerfcx(arg_erfcx) 
				log_curr_prob = - arg_erfcx**2 + log(erfcx_curr) - log_2  - &
												log(1.0 + sign(1.0_8,gamma_) * exp(-alpha_minus**2 + log(abs(gamma_)) - log_2 ))
			else
				erfcx_curr = myerfcx(-arg_erfcx) 
				log_curr_prob = log(1.0 - exp(log(erfcx_curr) - &
												log_2 - arg_erfcx**2))  - &
												log(1.0 + sign(1.0_8,gamma_) * exp(-alpha_minus**2 + log(abs(gamma_)) - log_2 ))
			end if
			curr_prob = 1.0 - exp(log_curr_prob)
		end if
			drawn_prob = oor_unit_dist(curr_prob,N_OOR)
	else ! normal Gibbs sampler
		call random_number(drawn_prob)
	end if

	!! Part 1, check left side of distribution
	u = log(drawn_prob) + log_2 - (b*c)/sqrt_a**2 + &
			log(1.0 + sign(1.0_8,gamma_) * exp(-alpha_plus**2 + log(abs(gamma_)) - log_2 + (b*c)/sqrt_a**2 ))
	if(u < -680.0 .and. max( abs(log(alpha_plus)) , abs(log(-alpha_minus)) ) > 3.0) then
		z = erfcinvexp_app(u)
	else if(max(abs(log(alpha_plus)) , abs(log(-alpha_minus)) ) > 3.0) then
		z = myerfcinv(exp(u))
	else
		z = myerfcinv(drawn_prob * (2.0 * exp(-(b*c)/sqrt_a**2) + exp(-alpha_plus**2) * gamma_ ))
	end if
	y = -(z - alpha_plus)/sqrt_a
	
	!! Part 2, check right side of distribution if y > 0
	if(y > 0.0) then
		u = log(1.0-drawn_prob) + log_2 + &
				log(1.0 + sign(1.0_8,gamma_) * exp( -alpha_minus**2 + log(abs(gamma_)) - log_2))
		if(u < -680.0 .and. abs(log(-alpha_minus)) > 3.0) then
			z = erfcinvexp_app(u)
		else if(abs(log(-alpha_minus)) > 3.0) then
			z = myerfcinv(exp(u))
		else
			z = myerfcinv((1.0-drawn_prob)*(2.0 +  exp(-alpha_minus**2) * gamma_))
		end if
		y =  (z - alpha_minus)/sqrt_a
	end if

end if
L1SamplerFunc = y
end function L1SamplerFunc








function erfcinvexp_app (u)
! approximates z = erfcinv(exp(u)) 
! use approximation : http://dlmf.nist.gov/7.17
implicit none

integer, parameter :: dp = kind(0.0d0)
real(kind=dp) u,erfcinvexp_app,theta,v,w,a2,a3,a4 

theta = - 1.144729885849400 - log(-u)
v = -theta - 2.0
w = 2.0/(theta - 2.0 * u)
a2 = 0.125 * v
a3 = - 0.03125 * (v**2 + 6.0*v - 6.0)
a4 = 1.0/384.0 * (4.0*v**3 + 27*v**2 + 108.0*v - 300.0)
erfcinvexp_app = 1.0/sqrt(w) + a2 * w**1.5 + a3 * w**2.5 + a4 * w**3.5

end function erfcinvexp_app










function oor_unit_dist(curr_prob,N_OOR)
! oriented overrelaxation of  curr_prob w.r.t.  uniform distribution on
! [0,1] and N_OOR
implicit none

integer*4 :: N_OOR,i,pos_curr
integer, parameter :: dp = kind(0.0d0)
real(kind=dp) :: curr_prob,oor_unit_dist
real(kind=dp),dimension(N_OOR+1) :: probs
integer*4,dimension(N_OOR+1) :: ind,ind_inv

ind =(/(i,i=1,N_OOR+1)/) 
ind_inv = ind

call random_number(probs)
probs(N_OOR+1) = curr_prob
call Insertion_Sort_Real(probs,ind,N_OOR+1)
call Insertion_Sort_Int4(ind,ind_inv,N_OOR+1)
pos_curr = ind_inv(N_OOR+1);
oor_unit_dist = probs(N_OOR + 2 - pos_curr)

end function oor_unit_dist







subroutine Insertion_Sort_Real(a,ind,N)
  implicit none
  integer*4 :: i, j,N
  integer, parameter :: dp = kind(0.0d0)
  real(kind = dp) :: a(N)
  integer*4 :: ind(N)
  real(kind=dp) :: temp
  
  do i = 2, size(a)
     j = i - 1
    temp = a(i)
     do while (j>=1 .and. a(j)>temp)
        a(j+1) = a(j)
        ind(j+1) = ind(j)
        j = j - 1
     end do
     a(j+1) = temp
     ind(j+1) = i
  end do
end subroutine Insertion_Sort_Real

subroutine Insertion_Sort_Int4(a,ind,N)
  implicit none
  integer*4 :: i, j, temp,N
  integer*4 a(N),ind(N)
 
  do i = 2, size(a)
     j = i - 1
     temp = a(i)
     do while (j>=1 .and. a(j)>temp)
        a(j+1) = a(j)
        ind(j+1) = ind(j)
        j = j - 1
     end do
     a(j+1) = temp
     ind(j+1) = i
  end do
  
end subroutine Insertion_Sort_Int4








  





function myerfcinv (arg)

implicit none

integer, parameter :: dp = kind(0.0d0)
real (kind=dp) arg, myerfcinv, myerfcx
real (kind=dp) p,p_low,p_high
real (kind=dp) a1,a2,a3,a4,a5,a6
real (kind=dp) c1,c2,c3,c4,c5,c6
real (kind=dp) b1,b2,b3,b4,b5
real (kind=dp) d1,d2,d3,d4
real (kind=dp) z,q,r, e, u
real (kind=dp) inv_sqrt_two, sqrt_two_pi, inf, neg_inf
  
data a1 /-3.969683028665376d01/
data a2 /2.209460984245205d02/
data a3 /-2.759285104469687d02/
data a4 /1.383577518672690d02/
data a6 /2.506628277459239d00/
data a5 /-3.066479806614716d01/
data b1 /-5.447609879822406d01/
data b2 /1.615858368580409d02/
data b3 /-1.556989798598866d02/
data b4 /6.680131188771972d01/
data c1 /-7.784894002430293d-03/
data b5 /-1.328068155288572d01/
data c2 /-3.223964580411365d-01/
data c3 /-2.400758277161838d00/
data c4 /-2.549732539343734d00/
data c5 /4.374664141464968d00/
data c6 /2.938163982698783d00/
data d1 /7.784695709041462d-03/
data d2 /3.224671290700398d-01/
data d3 /2.445134137142996d00/
data d4 /3.754408661907416d00/
data p_low /0.02425/
data inv_sqrt_two /0.707106781186547/
data sqrt_two_pi /2.506628274631000/
data inf /z'7f800000'/
data neg_inf /z'7f7fffff'/
  
p = arg/2  
p_high=1-p_low
  
! lower region  
if ( p <= p_low) then
	q=sqrt(-2*log(p))
		z=(((((c1*q+c2)*q+c3)*q+c4)*q+c5)*q+c6)/((((d1*q+d2)*q+d3)*q+d4)*q+1)
! central region  
else if ( p <= p_high) then
	q=p-0.5
	r=q*q
	z=(((((a1*r+a2)*r+a3)*r+a4)*r+a5)*r+a6)*q/(((((b1*r+b2)*r+b3)*r+b4)*r+b5)*r+1)
! upper region  
else
	q= sqrt(-2*log(1-p))
	z=-(((((c1*q+c2)*q+c3)*q+c4)*q+c5)*q+c6)/((((d1*q+d2)*q+d3)*q+d4)*q+1)
end if

! The relative error of the approximation has absolute value less
! than 1.15e-9.  One iteration of Halley's rational method (third
! order) gives full machine precision.
!  if( z < 0) then
e = myerfcx( -z * inv_sqrt_two)
u = sqrt_two_pi * (0.5 * e - p * exp((z*z)/2));      !f(z)/df(z)
!  else
!  u = sqrt_two_pi * (-0.5 * e  + (1- p) * exp(z**2/2));
!  call myerfcx ( z * inv_sqrt_two, e)
!  end if
!  z = z - u 							   			! Newtons method
z = z - u/( 1 + z *u /2 )           !Halley's method
      
myerfcinv = - z *  inv_sqrt_two
end function myerfcinv









function myerfcx(arg)

!*****************************************************************************80
!  Discussion:
!
!    This routine evaluates exp(x*x)*erfc(x)
!    for a real argument x.
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license.
!
!  Modified:
!
!    03 April 2007
!
!  Author:
!
!    Original FORTRAN77 version to compute erf, erfc and erfcx by William Cody.
!    FORTRAN90 version of this by John Burkardt.
!    This code is an extraction of the erfcx part by Felix Lucka
!
!  Reference:
!
!    William Cody,
!    Rational Chebyshev Approximations for the Error Function,
!    Mathematics of Computation,
!    Volume 23, Number 107, July 1969, pages 631-638.
!
!  Parameters:
!
!    Input, real ( kind = 8 ) ARG, the argument. The argument
!    must lie between XNEG and XMAX.
!
!    Output, real ( kind = 8 ) myerfcx, the value of the function,
!    which depends on the input value of JINT:
!
!
  implicit none

  real ( kind = 8 ) a(5)
  real ( kind = 8 ) arg, myerfcx
  real ( kind = 8 ) b(4)
  real ( kind = 8 ) c(9)
  real ( kind = 8 ) d(8)
  real ( kind = 8 ) del
  real ( kind = 8 ) four
  real ( kind = 8 ) half
  integer ( kind = 4 ) i
  real ( kind = 8 ) one
  real ( kind = 8 ) p(6)
  real ( kind = 8 ) q(5)
  real ( kind = 8 ) myerfcx
  real ( kind = 8 ) sixten
  real ( kind = 8 ) sqrpi
  real ( kind = 8 ) two
  real ( kind = 8 ) thresh
  real ( kind = 8 ) x
  real ( kind = 8 ) xbig
  real ( kind = 8 ) xden
  real ( kind = 8 ) xhuge
  real ( kind = 8 ) xinf
  real ( kind = 8 ) xmax
  real ( kind = 8 ) xneg
  real ( kind = 8 ) xnum
  real ( kind = 8 ) xsmall
  real ( kind = 8 ) y
  real ( kind = 8 ) ysq
  real ( kind = 8 ) zero
!
!  Mathematical constants
!
  data four /4.0d0/
  data one /1.0d0/
  data half /0.5d0/
  data two /2.0d0/
  data zero /0.0d0/
  data sqrpi /5.6418958354775628695d-1/
  data thresh /0.46875d0/
  data sixten /16.0d0/
!
!  Machine-dependent constants
!
  data xinf /1.79d308 /
  data xneg / -26.628d0 /
  data xsmall /1.11d-16/
  data xbig /26.543d0 /
  data xhuge /6.71d7/
  data xmax /2.53d307/
!
!  Coefficients for approximation to  erf  in first interval
!
  data a/3.16112374387056560d00,1.13864154151050156d02, &
         3.77485237685302021d02,3.20937758913846947d03, &
         1.85777706184603153d-1/
  data b/2.36012909523441209d01,2.44024637934444173d02, &
         1.28261652607737228d03,2.84423683343917062d03/
!
!  Coefficients for approximation to  erfc  in second interval
!
  data c/5.64188496988670089d-1,8.88314979438837594d0, &
         6.61191906371416295d01,2.98635138197400131d02, &
         8.81952221241769090d02,1.71204761263407058d03, &
         2.05107837782607147d03,1.23033935479799725d03, &
         2.15311535474403846d-8/
  data d/1.57449261107098347d01,1.17693950891312499d02, &
         5.37181101862009858d02,1.62138957456669019d03, &
         3.29079923573345963d03,4.36261909014324716d03, &
         3.43936767414372164d03,1.23033935480374942d03/
!
!  Coefficients for approximation to  erfc  in third interval
!
  data p/3.05326634961232344d-1,3.60344899949804439d-1, &
         1.25781726111229246d-1,1.60837851487422766d-2, &
         6.58749161529837803d-4,1.63153871373020978d-2/
  data q/2.56852019228982242d00,1.87295284992346047d00, &
         5.27905102951428412d-1,6.05183413124413191d-2, &
         2.33520497626869185d-3/

  x = arg
  y = abs ( x )
!
!  Evaluate erf for |X| <= 0.46875.
!
  if ( y <= thresh ) then

    ysq = zero
    if ( xsmall < y ) then
      ysq = y * y
    end if

    xnum = a(5) * ysq
    xden = ysq

    do i = 1, 3
      xnum = ( xnum + a(i) ) * ysq
      xden = ( xden + b(i) ) * ysq
    end do

    myerfcx = x * ( xnum + a(4) ) / ( xden + b(4) )

    myerfcx = exp ( ysq ) * (one - myerfcx)

    return
!
!  Evaluate erfc for 0.46875 <= |X| <= 4.0.
!
   else if ( y <= four ) then

     xnum = c(9) * y
     xden = y

     do i = 1, 7
       xnum = ( xnum + c(i) ) * y
       xden = ( xden + d(i) ) * y
     end do

     myerfcx = ( xnum + c(8) ) / ( xden + d(8) )

!
!  Evaluate erfc for 4.0 < |X|.
!
   else

     myerfcx = zero

     if ( xbig <= y ) then

       if ( xhuge <= y ) then
         myerfcx = sqrpi / y
         go to 300
       end if

     end if

     ysq = one / ( y * y )
     xnum = p(6) * ysq
     xden = ysq
     do i = 1, 4
       xnum = ( xnum + p(i) ) * ysq
       xden = ( xden + q(i) ) * ysq
      end do

      myerfcx = ysq * ( xnum + p(5) ) / ( xden + q(5) )
      myerfcx = ( sqrpi -  myerfcx ) / y

  end if
!
!  Fix up for negative argument, erf, etc.
!
  300 continue

    if ( x < zero ) then

      if ( x < xneg ) then
        myerfcx = xinf
      else
        ysq = aint ( x * sixten ) / sixten
        del = ( x - ysq ) * ( x + ysq )
        y = exp ( ysq * ysq ) * exp ( del )
        myerfcx = ( y + y ) - myerfcx
      end if

    end if

end function myerfcx




function DoubleExpSampler (c,N_O,x_curr)

	implicit none
	
	integer*4 N_O, pos_curr,i
	integer, parameter :: dp = kind(0.0d0)
! input/output
	real(kind=dp) :: c, x_curr, DoubleExpSampler
	real(kind=dp), dimension(N_O) :: x, z
	real(kind=dp), dimension(N_O+1) :: y
	integer*4,dimension(N_O+1) :: ind,ind_inv
	logical OOR_fl
	
	OOR_fl = (N_O > 1)

	if(OOR_fl) then !Use oriented over relaxation
		ind =(/(i,i=1,N_O+1)/) 
		ind_inv = ind

		call random_number(x)
		call random_number(z)
		y(1:N_O) = sign(log(x)/c,z-0.5) 
      y(N_O+1) = x_curr
      call Insertion_Sort_Real(y,ind,N_O+1)
		call Insertion_Sort_Int4(ind,ind_inv,N_O+1)
      pos_curr = ind_inv(N_O+1)
      DoubleExpSampler = y(N_O + 2 - pos_curr)
	else 
		call random_number(x)
		call random_number( y)
		DoubleExpSampler = sum(sign(log(x)/c , y-0.5))
	end if
            		
end function DoubleExpSampler




subroutine marsaglia_polar_method(n, y)
! draw standart normal distributed random variables by the 
! Marsaglia polar method (see Wikipedia)
implicit none

integer, parameter :: dp = kind(0.0d0)
!integer, parameter :: dp = selected_real_kind(p=15)
integer*8 :: i,n
real(kind=dp) :: y(n), unit_rn(2)
real(kind=dp) :: s, log_s, aux

do i = 1,n,2
!generate two new random variables
do 
	call random_number(unit_rn)
	unit_rn = 2.0 * unit_rn - 1.0
	s = unit_rn(1)*unit_rn(1) + unit_rn(2)*unit_rn(2)
	if(s < 1) exit
end do
aux = sqrt(-2.0*log(s)/s)
! assign them to the 
y(i) = unit_rn(1) * aux
if(i <= n) then
	y(i+1) = unit_rn(2) * aux
end if
end do


return
end subroutine marsaglia_polar_method


function crop_mat_ind(A)
! returns the indizes to crop the zero parts from A
implicit none

real*8 :: A(:,:)
integer*4 :: crop_mat_ind(4)
logical, allocatable :: any_row(:), any_col(:)
integer*4 :: m, n, i, i_min, i_max, j, j_max, j_min
m = size(A,1)
n = size(A,2)

allocate(any_row(m), any_col(n))
any_col = any(abs(A) > 0.0,1)
any_row = any(abs(A) > 0.0,2)

do i=1,m
	if(any_row(i)) then
		i_min = i
		exit
	end if
end do
do i=m,1,-1
	if(any_row(i)) then
		i_max = i
		exit
	end if
end do

do j=1,n
	if(any_col(j)) then
		j_min = j
		exit
	end if
end do
do j=n,1,-1
	if(any_col(j)) then
		j_max = j
		exit
	end if
end do

crop_mat_ind = (/ i_min,i_max,j_min,j_max/)

end function crop_mat_ind


function pad_array(A,padsize)
! returns a padded matrix
implicit none

real*8 :: A(:,:)
real*8, allocatable, dimension(:,:) :: pad_array
integer*4 :: m,n,m_p,n_p,padsize(2),i,j

m = size(A,1)
n = size(A,2)
m_p = m + 2*padsize(1)
n_p = n + 2*padsize(2)

allocate(pad_array(m_p,n_p))

pad_array = 0.0

do i=1,m
	do j=1,n
		pad_array(i+padsize(1),j+padsize(2)) = A(i,j)
	end do
end do


end function pad_array


subroutine my_pause(t_sec)

implicit none

real*8 t_sec,t_start,t_now

call cpu_time(t_start)

do
call cpu_time(t_now)
if(t_now-t_start >= t_sec) then
exit
end if
end do

end subroutine my_pause


